//2010 0.1 Copyright (C) Michael Niedermayer 1998

#include <stdlib.h>
#include <stdio.h>
#include <go32.h>
#include <dpmi.h>
#include <string.h>
#include <sys/nearptr.h>
#include <sys/farptr.h>
#include "2_all.h"
#include "2_hw_mem.h"
#include "2010.h"

 extern volatile xxx dispbuf, last_dispbuf, grab_dispbuf;
 extern volatile LockList  *locklist;
 extern volatile AllocList *alloclist;
 extern volatile VdsList *vdslist;
 extern volatile MapList *maplist;

 dds g_vds_info={size: 0, offset: 0, selector: 0, id: 0, phys: 0};

void locklin(void *start, void *end){
 volatile LockList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=new LockList;
 llp2=locklist;
 while(llp2!=NULL && llp2->next!=NULL) llp2=llp2->next;

 meminfo.address=(unsigned long)(start);
 meminfo.size=(unsigned long)(end - start);
 if(__dpmi_lock_linear_region(&meminfo)!=0) error(Lock);

 if(llp2!=NULL) llp2->next=llp;
 else locklist=llp2=llp;
 llp->start=start;
 llp->end=end;
 llp->next=NULL;
}

void unlocklin(void *start, void *end){
 volatile LockList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=locklist;
 while(llp!=NULL && (llp->start!=start || llp->end!=end)) llp=llp->next;
 if(llp==NULL){
   while(llp!=NULL && ((llp->start!=start && llp->end!=end)
                    || llp->start>start  || llp->end<end)) llp=llp->next;
 }
 if(llp==NULL) error(error_code(-8));

 meminfo.address=(u_long)(start);
 meminfo.size=(u_long)(end - start);
 __dpmi_unlock_linear_region(&meminfo);

 if(llp->start<start){
   llp->end=start;
 }else if(llp->end>end){
   llp->start=end;
 }else{
   llp2=locklist;
   while(llp2!=NULL && llp2->next!=llp) llp2=llp2->next;
   if(llp2!=NULL){
     llp2->next=llp->next;
   }else{
     locklist=llp->next;
   }
   delete llp;
 }
}

void lock(void *start, void *end){
 locklin(start + __djgpp_base_address, end + __djgpp_base_address);
}

void unlock(void *start, void *end){
 unlocklin(start + __djgpp_base_address, end + __djgpp_base_address);
}

void unlockall(void){
 volatile LockList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=locklist;
 while(llp!=NULL){
   meminfo.address=(u_long)(llp->start);
   meminfo.size=(u_long)(llp->end - llp->start);
   __dpmi_unlock_linear_region(&meminfo);
   llp2=llp;
   llp=llp->next;
   delete llp2;
 }
}

void dpmi_addlinlock(volatile AllocList *old, volatile AllocList **al){
 volatile AllocList *llp, *llp2;

 llp=new AllocList;
 llp2=*al;
 while(llp2!=NULL && llp2->next!=NULL && llp2->next->phys<old->phys) llp2=llp2->next;

 if(*al!=NULL && (*al)->phys>old->phys) llp2=NULL;

 *llp=*old;
 if(llp2==NULL){
   llp->next=*al;
   *al=llp;
 }else{
   llp->next=llp2->next;  
   llp2->next=llp;
 }
}

u_long dpmi_alloclinlock(volatile AllocList **al){
 volatile AllocList *llp, *llp2;
 __dpmi_meminfo meminfo;
 u_long phys;

 meminfo.size=4096;
 if(__dpmi_allocate_memory(&meminfo)==-1) return 0;
 if(__dpmi_lock_linear_region(&meminfo)==-1){
   __dpmi_free_memory(meminfo.handle);
   return 0;
 }

 if((meminfo.address & 4095) != 0) error(error_code(-90));

 phys=getphys(meminfo.address);

 llp=(AllocList *)(meminfo.address - __djgpp_base_address);
 llp2=*al;
 while(llp2!=NULL && llp2->next!=NULL && llp2->next->phys<phys) llp2=llp2->next;

 if(*al!=NULL && (*al)->phys>phys) llp2=NULL;

 if(llp2==NULL){
   llp->next=*al;
   *al=llp;
 }else{
   llp->next=llp2->next;
   llp2->next=llp;
 }
 llp->addr=meminfo.address;
 llp->hand=meminfo.handle;
 llp->phys=phys;

 return meminfo.address;
}

void dpmi_freelinlock(volatile AllocList *old, volatile AllocList **al){
 volatile AllocList *llp2;
 __dpmi_meminfo meminfo;


 meminfo.address=old->addr;
 meminfo.size=4096;
 meminfo.handle=old->hand;

 llp2=*al;
 while(llp2!=NULL && llp2->next!=old) llp2=llp2->next;
 if(llp2!=NULL){
   llp2->next=old->next;
 }else{
   *al=old->next;
 }

 if((u_long)old + __djgpp_base_address != meminfo.address) delete old;

 if(meminfo.handle!=0){
   __dpmi_unlock_linear_region(&meminfo);
   __dpmi_free_memory(meminfo.handle);
 }

}

void dpmi_freeall(volatile AllocList **al){
 volatile AllocList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=*al;
 while(llp!=NULL){
   meminfo.address=llp->addr;
   meminfo.size=4096;
   meminfo.handle=llp->hand;
   llp2=llp;
   llp=llp->next;
   if((u_long)llp2 + __djgpp_base_address != meminfo.address) delete llp2;
   if(meminfo.handle!=0){
     __dpmi_unlock_linear_region(&meminfo);
     __dpmi_free_memory(meminfo.handle);
   }
 }
 *al=NULL;
}


xxx alloc_cont(int size){
 int i=0;
 xxx ret;
 u_long lphys, addr, maxaddr;
 int count, maxcount;
 volatile AllocList *llp, *llp2, *temp_al=NULL;
 __dpmi_meminfo meminfo;
 int length;
 dds vds_info;

 size= (size + 4095) & (~4095);     //round up to next 4096 boundray

 checkVDS();

 
 addr=(u_long)map(2048*1024, 2048*1024 + 1024*1024);  //check if mapping of RAM is ok
 unmap((void*)addr);

 if(g_vds_info.phys==0 && g_vds_info.size==0){
   vds_info=vds_alloc(size<<1);
   if(vds_info.phys==0xFFFFFFFF)
     vds_info=vds_alloc(size);
   else
     g_vds_info=vds_info;
 }
 else{
   vds_info=g_vds_info;
   vds_info.phys+=size;
 }
 if(vds_info.phys==0xFFFFFFFF){
   length=maxcount=maxaddr=0;

   while(size>maxcount){
     if(dpmi_alloclinlock(&temp_al)==0){
       dpmi_freeall(&temp_al);
       error(MemAlloc);
     }
  
     llp=temp_al;
     lphys=count=addr=0;
     while(llp!=NULL){
       if((lphys + 4096) == llp->phys && (llp->phys & (1024*1024*4-1)) != 0)
         count+=4096;
       else{
         count=0;
         addr=llp->phys;
       }
       if(count>maxcount){
         maxcount=count;
         maxaddr=addr;
       }
       lphys=llp->phys;
  
       llp=llp->next;
     }
     length+=4096;
     printf("Allocated %d, MaxCont %d, at Address %lX\r",length , maxcount, maxaddr);
     fflush(stdout);
   }
   printf("\n");
  
   llp=temp_al;
   while(llp!=NULL){
     llp2=llp;
     llp=llp->next;
     if(llp2->phys<maxaddr){
       dpmi_freelinlock(llp2, &temp_al);
     }
     else if(llp2->phys>maxaddr+maxcount){
       dpmi_freelinlock(llp2, &temp_al);
     }
     else{
       dpmi_addlinlock(llp2, &alloclist);
     }
   }
 }else{
   maxaddr=vds_info.phys;
   printf("AllocatedContVDS %d, at Address %lX\n", size , maxaddr);
   fflush(stdout);
 }
 meminfo.size=size;
 meminfo.address=maxaddr;
 ret.b=(byte*)(map(maxaddr, maxaddr+size) - __djgpp_base_address);

 memset(ret.b, 0, size);

 i=~(1024*1024*4-1);
 if( (maxaddr & i) != ((maxaddr+size) & i) ) error(MemAlloc);
 //FIX ME (should be retried until ok)
 ret.phys=maxaddr;
 return ret;

}
                                   
void checkVDS(void){               

 if((_farpeekb(_dos_ds, 0x47B) & 32) != 32) error(NoVDS);

}

u_long getphys(u_long addr){
__dpmi_regs r;
dds info;

 info.selector=0;
 info.size=4096;
 info.offset=addr;

 dosmemput(&info, sizeof(dds), __tb); // store in buffer

 r.x.ax=0x8103;
 r.x.dx=4;
 r.x.es = __tb >> 4;
 r.x.di = 0;
 __dpmi_int(0x4B, &r);
 if((r.x.flags & carry_flag) == carry_flag) error(VDS);
 dosmemget(__tb, sizeof(dds), &info);

 r.x.ax=0x8104;
 r.x.dx=0;
 r.x.es = __tb >> 4;
 r.x.di = 0;
 __dpmi_int(0x4B, &r);
 if((r.x.flags & carry_flag) == carry_flag) error(VDS);

 return info.phys;

}

dds vds_alloc(int size){
 dds vds_info;
 __dpmi_regs r;
 volatile VdsList *vd;

 vds_info.size=size;

 dosmemput(&vds_info, sizeof(dds), __tb); // store in buffer

 r.x.ax=0x8107;
 r.x.dx=0;
 r.x.es=__tb >> 4;
 r.x.di=0;

 __dpmi_int(0x4B, &r);
 if((r.x.flags & carry_flag) == carry_flag){
   vds_info.phys=0xFFFFFFFF;
   return vds_info;
 };
 dosmemget(__tb, sizeof(dds), &vds_info);

 vd=vdslist;
 while(vd!=NULL && vd->next!=NULL) vd=vd->next;
 if(vd==NULL){
   vdslist=vd=new VdsList;
 }else{
   vd->next=new VdsList;
   vd=vd->next;
 }

 vd->vds_info=vds_info;
 vd->next=NULL;

 return vds_info;
}

void vds_free(dds){

 error( error_code(-99) );
 // fix me

}

void vds_freeall(void){
 __dpmi_regs r;
 volatile VdsList *vd, *vd2;

 vd=(VdsList*)vdslist;
 while(vd!=NULL){
   dosmemput((dds*)(&vd->vds_info), sizeof(dds), __tb); // store in buffer
   r.x.ax=0x8108;
   r.x.dx=0;
   r.x.es=__tb >> 4;
   r.x.di=0;

   __dpmi_int(0x4B, &r);
   vd2=vd;
   vd=vd->next;
   delete vd2;
 }

 vdslist=NULL;
}

void *map(u_long start, u_long end){
 volatile MapList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=new MapList;
 llp2=maplist;
 while(llp2!=NULL && llp2->next!=NULL) llp2=llp2->next;

 meminfo.address=start;
 meminfo.size=end - start;
 if(__dpmi_physical_address_mapping(&meminfo)!=0) error(Map);

 if(llp2!=NULL) llp2->next=llp;
 else maplist=llp2=llp;
 llp->start=(void*)meminfo.address;
 llp->next=NULL;
 return (void*)meminfo.address;
}

void unmap(void *start){
 volatile MapList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=maplist;
 while(llp!=NULL && llp->start!=start) llp=llp->next;
 if(llp==NULL) error(error_code(-19));

 meminfo.address=(u_long)(start);
 __dpmi_free_physical_address_mapping(&meminfo);

 llp2=maplist;
 while(llp2!=NULL && llp2->next!=llp) llp2=llp2->next;
 if(llp2!=NULL){
   llp2->next=llp->next;
 }else{
   maplist=llp->next;
 }
 delete llp;
}

void unmapall(void){
 volatile MapList *llp, *llp2;
 __dpmi_meminfo meminfo;

 llp=maplist;
 while(llp!=NULL){
   meminfo.address=(u_long)(llp->start);
   __dpmi_free_physical_address_mapping(&meminfo);
   llp2=llp;
   llp=llp->next;
   delete llp2;
 }
}
